-- -----------------------------------
-- Create Ace3 addon instances
-- -----------------------------------

MasterLoot = LibStub("AceAddon-3.0"):NewAddon("MasterLoot", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0", "AceHook-3.0", "AceConsole-3.0")
local AceGUI = LibStub("AceGUI-3.0")

-- -----------------------------------
-- Ace3 Standard Methods
-- -----------------------------------

function MasterLoot:OnInitialize()
 -- Code that you want to run when the addon is first loaded goes here.
	self.defaults = {
		char = {
			drop = {},
			player = {},
			playerGuids = {},
			playerIDs = {},
			slotPlayerItem = {
				head = {},
				neck = {},
				shoulder = {},
				back = {},
				chest = {},
				wrist = {},
				hands = {},
				waist = {},
				legs = {},
				feet = {},
				finger0 = {},
				finger1 = {},
				trinket0 = {},
				trinket1 = {},
				mainHand = {},
				secondaryHand = {}
			}
		},
	}
	self.fk = {}
	self.triggers = {}
	self.inspection = {
		Q = {}
	}

	self.db = LibStub("AceDB-3.0"):New(self:GetName() .. "DB", self.defaults)
	if self.IsTest then
		self.db:ResetProfile(false, false)
	end
	if not self:IsHooked("NotifyInspect") then self:Hook("NotifyInspect", "NotifyInspect", true) end
	-- Register foreign keys for cascade delete
	self:RegisterFK(self.db.char.playerGuids, nil, self.db.char.player)
	self:RegisterFK(self.db.char.playerIDs, nil, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.head, self.db.char.slotPlayerItem.head, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.neck, self.db.char.slotPlayerItem.neck, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.shoulder, self.db.char.slotPlayerItem.shoulder, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.back, self.db.char.slotPlayerItem.back, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.chest, self.db.char.slotPlayerItem.chest, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.wrist, self.db.char.slotPlayerItem.wrist, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.hands, self.db.char.slotPlayerItem.hands, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.waist, self.db.char.slotPlayerItem.waist, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.legs, self.db.char.slotPlayerItem.legs, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.feet, self.db.char.slotPlayerItem.feet, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.finger0, self.db.char.slotPlayerItem.finger0, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.finger1, self.db.char.slotPlayerItem.finger1, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.trinket0, self.db.char.slotPlayerItem.trinket0, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.trinket1, self.db.char.slotPlayerItem.trinket1, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.mainHand, self.db.char.slotPlayerItem.mainHand, self.db.char.player)
	self:RegisterFK(self.db.char.slotPlayerItem.secondaryHand, self.db.char.slotPlayerItem.secondaryHand, self.db.char.player)
	-- Register triggers
	self:RegisterTrigger(self.db.char.player, "BEFORE", "DELETE", self.TriggerPlayerBeforeDelete)
	self:RegisterTrigger(self.db.char.player, "AFTER", "INSERT", self.TriggerPlayerAfterInsert)
	self:RegisterTrigger(self.db.char.player, "AFTER", "UPDATE", self.TriggerPlayerAfterUpdate)
	for slot, playerInfo in pairs(self.db.char.slotPlayerItem) do
		self:RegisterTrigger(self.db.char.slotPlayerItem[slot], "AFTER", "INSERT", self.TriggerSlotPlayerItemAfterInsert)
		self:RegisterTrigger(self.db.char.slotPlayerItem[slot], "AFTER", "UPDATE", self.TriggerSlotPlayerItemAfterUpdate)
		self:RegisterTrigger(self.db.char.slotPlayerItem[slot], "AFTER", "DELETE", self.TriggerSlotPlayerItemAfterDelete)
	end
	self:RegisterTrigger(self.db.char.drop, "AFTER", "INSERT", self.TriggerDropAfterInsert)
	self:RegisterTrigger(self.db.char.drop, "AFTER", "UPDATE", self.TriggerDropAfterUpdate)
	self:LoadScales()
	self:AddSocketsToScales()
	self:CacheAllItems()
end

function MasterLoot:OnEnable()
    -- Called when the addon is enabled
	self.timers = {
		inspection = {},
	}
	self:RegisterFK(self.timers.inspection, self.timers.inspection, self.db.char.playerGuids)
	self.playerName = GetUnitName("player", true)
	self.playerGuid = UnitGUID("player")
	self.lootMethod, self.masterLooterPartyID, self.masterLooterRaidID = GetLootMethod()
	self.lootThreshold = GetLootThreshold()
	self:DBUpsert(self.db.char.player, {[self.playerName] = {name = self.playerName, id = "player", guid = self.playerGuid, class = select(2, UnitClass("player"))}})
	-- Register event handlers
	if self.IsTest then
		self:RegisterMessage("PLAYER_ENTERING_WORLD", "GROUP_ROSTER_UPDATE")
		self:RegisterMessage("GROUP_JOINED", "GROUP_ROSTER_UPDATE")
		self:RegisterMessage("GROUP_ROSTER_UPDATE", "GROUP_ROSTER_UPDATE")
		self:RegisterMessage("PARTY_LEADER_CHANGED", "GROUP_ROSTER_UPDATE")
		self:RegisterMessage("PARTY_CONVERTED_TO_RAID", "GROUP_ROSTER_UPDATE")
		self:RegisterMessage("PARTY_LOOT_METHOD_CHANGED", "PARTY_LOOT_METHOD_CHANGED")
		self:RegisterBucketMessage("UNIT_INVENTORY_CHANGED", 2, "UNIT_INVENTORY_CHANGED")
		self:RegisterMessage("INSPECT_READY", "INSPECT_READY")
		self:RegisterMessage("PLAYER_SPECIALIZATION_CHANGED", "NotifyInspect")
		self:RegisterMessage("GET_ITEM_INFO_RECEIVED", "GET_ITEM_INFO_RECEIVED")
		self:RegisterMessage("LOOT_OPENED", "LOOT_OPENED")
		self:RegisterMessage("CHAT_MSG_LOOT", "CHAT_MSG_LOOT")
	else
		self:RegisterEvent("PLAYER_ENTERING_WORLD", "GROUP_ROSTER_UPDATE")
		self:RegisterEvent("GROUP_JOINED", "GROUP_ROSTER_UPDATE")
		self:RegisterEvent("GROUP_ROSTER_UPDATE", "GROUP_ROSTER_UPDATE")
		self:RegisterEvent("PARTY_LEADER_CHANGED", "GROUP_ROSTER_UPDATE")
		self:RegisterEvent("PARTY_CONVERTED_TO_RAID", "GROUP_ROSTER_UPDATE")
		self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED", "PARTY_LOOT_METHOD_CHANGED")
		self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", 2, "UNIT_INVENTORY_CHANGED")
		self:RegisterEvent("INSPECT_READY", "INSPECT_READY")
		self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED", "NotifyInspect")
		self:RegisterEvent("GET_ITEM_INFO_RECEIVED", "GET_ITEM_INFO_RECEIVED")
		self:RegisterEvent("LOOT_OPENED", "LOOT_OPENED")
		self:RegisterEvent("CHAT_MSG_LOOT", "CHAT_MSG_LOOT")
	end
	
	-- Cleanup and update SV tables
	self:UpdateAllRoster()
	for key, value in pairs(self.db.char.player) do
		self:NotifyInspect(key)
	end
	-- Cleanup drop table and reset timers
	local drops = {}
	for id, dropInfo in pairs(self.db.char.drop) do
		-- If expTime <= current time, remove
		if dropInfo.expTime <= time() then
			drops[id] = dropInfo
		end
	end
	self:DBDelete(self.db.char.drop, drops)
	self.PlayerUpgradeFrame = self:CreatePlayerUpgradeFrame()
	self.PlayerUpgradeFrame.frame:Hide()
	self:RegisterChatCommand("mlshow", function() self.PlayerUpgradeFrame.frame:Show() end, true) 
end

function MasterLoot:OnDisable()
    -- Called when the addon is disabled
	MasterLoot:CancelAllTimers()
	MasterLoot:UnregisterAllBuckets()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD", "GROUP_ROSTER_UPDATE")
	self:UnregisterEvent("GROUP_JOINED", "GROUP_ROSTER_UPDATE")
	self:UnregisterEvent("GROUP_ROSTER_UPDATE", "GROUP_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_LEADER_CHANGED", "GROUP_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_CONVERTED_TO_RAID", "GROUP_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_LOOT_METHOD_CHANGED", "PARTY_LOOT_METHOD_CHANGED")
	self:UnregisterEvent("INSPECT_READY", "INSPECT_READY")
	self:UnregisterEvent("PLAYER_SPECIALIZATION_CHANGED", "NotifyInspect")
	self:UnregisterEvent("GET_ITEM_INFO_RECEIVED", "GET_ITEM_INFO_RECEIVED")
	self:UnregisterEvent("LOOT_OPENED", "LOOT_OPENED")
	self:UnregisterEvent("CHAT_MSG_LOOT", "CHAT_MSG_LOOT")
end

-- -----------------------------------
-- Hooks
-- -----------------------------------

function MasterLoot:NotifyInspect(unit, restart)
	local name
	if restart or (not self.inspection.status)  then self.inspection.status = "idle" end
	if unit then
		name = GetUnitName(unit, true)
		local found = false
		for key, value in pairs(self.inspection.Q) do
			if value == name then found = true end
		end
		if not found then tinsert(self.inspection.Q, name) end
	end
	if self.inspection.status ~= "stopped" then
		self.inspection.status = "running"
		-- If there are requests in queue, process the next request
		if #self.inspection.Q > 0 then
			if (not self.inspection.timer) or self:TimeLeft(self.inspection.timer) == 0 then
				self.inspection.numRequests = 0
				self.inspection.timer = self:ScheduleRepeatingTimer(function () self.inspection.numRequests = 0 end, 2)
			end
			-- If we have already reached our limit for this interval, then wait for bandwidth to send one more and try again
			if self.inspection.numRequests == 1 then
				self:ScheduleTimer("NotifyInspect", 2)
			else
			-- We're not at the limit so go ahead
				self.inspection.numRequests = self.inspection.numRequests + 1
				if CanInspect(self.inspection.Q[1]) and ((not self.timers.inspection[guid]) or self:TimeLeft(self.timers.inspection[guid] == 0)) then
					local guid = UnitGUID(self.inspection.Q[1])
					self.inspection.status = "stopped"
					self.hooks.NotifyInspect(self.inspection.Q[1])
					-- schedule a retry timer in case of timeout
					self.timers.inspection[guid] = self:ScheduleTimer("NotifyInspect", self:round(2 + ((select(4, GetNetStats()) * 2) / 1000)), name, true)
				elseif self.db.char.player[name] then
					-- move to the back of the queue
					tinsert(self.inspection.Q, name)
				end
				tremove(self.inspection.Q, 1)
				-- Continue the process until no requests are left
				self:NotifyInspect()
			end
		else
			self.inspection.status = "idle"
		end
	end
end

-- -----------------------------------
-- DBMS functions
-- -----------------------------------

function MasterLoot:RegisterFK(fkobj, fk, pkobj)
	assert(fkobj, "Error in RegisterFK: no fkobj specified.")
	assert(pkobj, "Error in RegisterFK: no pkobj specified.")
	assert(type(fkobj) == "table", "Error in RegisterFK: fkobj not of type table.")
	if fk then assert(type(fk) == "string" or (type(fk) == "table" and fk == fkobj), "Error in RegisterFK: fk not of type string or equal to fkobj.") end
	assert(type(pkobj) == "table", "Error in RegisterFK: pkobj not of type table.")
	self.fk[tostring(fkobj) .. "|" .. tostring(fk)] = {fkobj = fkobj, fk = fk, pkobj = pkobj}
end

function MasterLoot:RegisterTrigger(obj, BeforeAfter, operation, funct)
	assert(obj, "Error in RegisterTrigger: object not specified.")
	assert(BeforeAfter, "Error in RegisterTrigger: Before or After not specified.")
	assert(operation, "Error in RegisterTrigger: operation not specified.")
	assert(funct, "Error in RegisterTrigger: function not specified.")
	assert(type(obj) == "table", "Error in RegisterTrigger: obj is not of type table.")
	assert(BeforeAfter == "BEFORE" or BeforeAfter == "AFTER", "Error in RegisterTrigger: BeforeAfter must be keyword BEFORE or AFTER.")
	assert(operation == "DELETE" or operation == "UPDATE" or operation == "INSERT", "Error in RegisterTrigger: invalid operation.")
	assert(type(funct) == "function", "Error in RegisterTrigger: function is not of type function.")
	if not self.triggers[obj] then
		self.triggers[obj] = {[BeforeAfter] = {[operation] = {funct}}}
	elseif not self.triggers[obj][BeforeAfter] then
		self.triggers[obj][BeforeAfter] = {[operation] = {funct}}
	elseif not self.triggers[obj][BeforeAfter][operation] then
		self.triggers[obj][BeforeAfter][operation] = {funct}
	else
		tinsert(self.triggers[obj][BeforeAfter][operation], funct)
	end
end

function MasterLoot:ProcessTrigger(obj, BeforeAfter, operation, deleted, inserted)
	assert(obj, "Error in ProcessTrigger: object not specified.")
	assert(BeforeAfter, "Error in ProcessTrigger: Before or After not specified.")
	assert(operation, "Error in ProcessTrigger: operation not specified.")
	assert(deleted, "Error in ProcessTrigger: deleted not specified.")
	assert(inserted, "Error in ProcessTrigger: inserted not specified.")
	assert(type(obj) == "table", "Error in ProcessTrigger: obj is not of type table.")
	assert(BeforeAfter == "BEFORE" or BeforeAfter == "AFTER", "Error in ProcessTrigger: BeforeAfter must be keyword BEFORE or AFTER.")
	assert(operation == "DELETE" or operation == "UPDATE" or operation == "INSERT", "Error in ProcessTrigger: obj is not of type table.")
	assert(type(deleted) == "table", "Error in ProcessTrigger: deleted is not of type table.")
	assert(type(inserted) == "table", "Error in ProcessTrigger: inserted is not of type table.")
	if self.triggers and self.triggers[obj] and self.triggers[obj][BeforeAfter] and self.triggers[obj][BeforeAfter][operation] then
		for key, f in pairs(self.triggers[obj][BeforeAfter][operation]) do
			deleted, inserted = f(self, obj, deleted, inserted)
		end
	end
	return deleted, inserted
end

function MasterLoot:DML(operation, obj, deleted, inserted)
	assert(operation, "Error in DML: operation not specified.")
	assert(obj, "Error in DML: object not specified.")
	assert(deleted, "Error in DML: deleted table not specified.")
	assert(inserted, "Error in DML: inserted table not specified.")
	assert(type(obj) == "table", "Error in DML: obj is not of type table.")
	assert(operation == "DELETE" or operation == "UPDATE" or operation == "INSERT", "Error in DML: invalid operation.")
	assert(type(deleted) == "table", "Error in DML: deleted is not of type table.")
	assert(type(inserted) == "table", "Error in DML: inserted is not of type table.")
	for k1, v1 in pairs(deleted) do
		local found = false
		for k2, v2 in pairs(inserted) do
			if k1 == k2 then found = true end
		end
		if not found then obj[k1] = nil end
	end
	for key, value in pairs(inserted) do
		obj[key] = value
	end
	self:ProcessTrigger(obj, "AFTER", operation, deleted, inserted)
end

function MasterLoot:DBDelete(obj, deleted)
	-- validate the parameters, that the object exists, and retrieve the full records
	assert(type(obj) == "table", "Error in DBDelete: object does not exist or is not of type table.")
	assert(type(deleted) == "table", "Error: in DBDelete: \"deleted\" does not exist or is not of type table.")
	-- process any possible cascade deletes
	if self.fk then
		for key, value in pairs(self.fk) do
			if value.pkobj == obj then
				local cascade = {}
				if not value.fk then
					for k, v in pairs(value.fkobj) do
						for k2, v2 in pairs(deleted) do
							if v == k2 then cascade[k] = v end
						end
					end
				elseif type(value.fk) == "table" then
					for k, v in pairs(value.fkobj) do
						for k2, v2 in pairs(deleted) do
							if k == k2 then cascade[k] = v end
						end
					end
				else
					for k, v in pairs(value.fkobj) do
						for k2, v2 in pairs(deleted) do
							if v[value.fk] == k2 then cascade[k] = v end
						end
					end
				end
				self:DBDelete(value.fkobj, cascade)
			end
		end
	end
	local errorstate = false
	local recordcount = 0
	local errors = {}
	local validated = {}
	for key, value in pairs(deleted) do
		local found = false
		for index, v in pairs(obj) do
			if key == index then
				found = true
				recordcount = recordcount + 1
				validated[key] = value
			end
		end
		if not found then
			errorstate = true
			tinsert(errors, key)
		end
	end
	if recordcount > 0 then
		validated, _ = self:ProcessTrigger(obj, "BEFORE", "DELETE", validated, {})
		self:DML("DELETE", obj, validated, {})
	end
	if errorstate then
		return false, recordcount, errors
	else
		return true, recordcount
	end
end

function MasterLoot:DBUpsert(obj, inserted)
	-- validate the parameters, that the object exists, and retrieve the full records
	assert(type(obj) == "table", "Error in DBUpdate: object does not exist or is not of type table.")
	assert(type(inserted) == "table", "Error: in DBUpdate: \"inserted\" does not exist or is not of type table.")
	local insertcount = 0
	local updatecount = 0
	local deleted = {}
	local updated = {}
	for key, value in pairs(inserted) do
		local found = false
		for k, v in pairs(obj) do
			if k == key then
				found = true
				updatecount = updatecount + 1
				deleted[k] = v
				updated[key] = value
				inserted[key] = nil
				-- if the value is a table, add missing key, value pairs to inserted
				if type(value) == "table" then
					for k2, v2 in pairs(v) do
						local found2 = false
						for k3, v3 in pairs(updated[key]) do
							if k2 == k3 then found2 = true end
						end
						if not found2 then updated[key][k2] = v2 end
					end
				end
			end
		end
		if not found then insertcount = insertcount + 1 end
	end
	if insertcount > 0 then
		_, inserted = self:ProcessTrigger(obj, "BEFORE", "INSERT", {}, inserted)
		self:DML("INSERT", obj, {}, inserted)
	end
	if updatecount > 0 then
		deleted, updated = self:ProcessTrigger(obj, "BEFORE", "UPDATE", deleted, updated)
		self:DML("UPDATE", obj, deleted, updated)
	end
	return true, updatecount, insertcount
end

function MasterLoot:DBInsert(obj, inserted)
	-- validate the parameters, that the object exists, and that existing records with the same key do not exist
	assert(type(obj) == "table", "Error in DBInsert: object does not exist or is not of type table.")
	assert(type(inserted) == "table", "Error: in DBInsert: \"inserted\" does not exist or is not of type table.")
	local errorstate = false
	local recordcount = 0
	local errors = {}
	local validated = {}
	for key, value in pairs(inserted) do
		local found = false
		for k, v in pairs(obj) do
			if k == key then
				found = true
				errorstate = true
				errors[key] = value
			end
		end
		if not found then
			recordcount = recordcount + 1
			validated[key] = value
		end
	end
	if recordcount > 0 then
		_, validated = self:ProcessTrigger(obj, "BEFORE", "INSERT", {}, validated)
		self:DML("INSERT", obj, {}, validated)
	end
	if errorstate then
		return false, recordcount, errors
	else
		return true, recordcount
	end
end

function MasterLoot:DBUpdate(obj, inserted)
	-- validate the parameters, that the object exists, and retrieve the full records
	assert(type(obj) == "table", "Error in DBUpdate: object does not exist or is not of type table.")
	assert(type(inserted) == "table", "Error: in DBUpdate: \"inserted\" does not exist or is not of type table.")
	local errorstate = false
	local recordcount = 0
	local errors = {}
	local validated = {}
	local deleted = {}
	for key, value in pairs(inserted) do
		local found = false
		for k, v in pairs(obj) do
			if k == key then
				found = true
				recordcount = recordcount + 1
				validated[key] = value
				deleted[k] = v
				-- if the value is a table, add missing key, value pairs to inserted
				if type(value) == "table" then
					for k2, v2 in pairs(deleted[k]) do
						local found2 = false
						for k3, v3 in pairs(validated[key]) do
							if k2 == k3 then found2 = true end
						end
						if not found2 then validated[key][k2] = v2 end
					end
				end
			end
		end
		if not found then
			errorstate = true
			errors[key] = value
		end
	end
	if recordcount > 0 then
		deleted, validated = self:ProcessTrigger(obj, "BEFORE", "UPDATE", deleted, validated)
		self:DML("UPDATE", obj, deleted, validated)
	end
	if errorstate then
		return false, recordcount, errors
	else
		return true, recordcount
	end
end

-- -----------------------------------
-- Triggers
-- -----------------------------------

function MasterLoot:TriggerPlayerBeforeDelete(obj, deleted, inserted)
	for key, value in pairs(deleted) do
		if self.timers.inspection[value.guid] then
			self:CancelTimer(self.timers.inspection[value.guid])
		end
	end
	return deleted, inserted
end

function MasterLoot:TriggerPlayerAfterInsert(obj, deleted, inserted)
	local ids = {}
	local guids = {}
	for key, value in pairs(inserted) do
		ids[value.id] = key
		guids[value.guid] = key
	end
	self:DBInsert(self.db.char.playerIDs, ids)
	self:DBInsert(self.db.char.playerGuids, guids)
	for key, value in pairs(inserted) do
		self:NotifyInspect(key)
	end
	return deleted, inserted
end

function MasterLoot:TriggerPlayerAfterUpdate(obj, deleted, inserted)
	local idDelete = {}
	local idUpsert = {}
	local playerUpdates = {}
	for k1, v1 in pairs(deleted) do
		for k2, v2 in pairs(inserted) do
			if k1 == k2 then
				-- If the unit ID changed then delete from payerIDs (unless the player ID is player) and add the new ID
				if v1.id ~= v2.id then
					for id, playerName in pairs(self.db.char.playerIDs) do
						if k1 == playerName and id ~= "player" then
							idDelete[id] = playerName
							idUpsert[v2.id] = k2
						end
					end
				end
				-- If currentSpec has changed, update usesIntellect
				if v2.class and v2.currentSpec and ((not v1.currentSpec) or v1.currentSpec ~= v2.currentSpec) then
					if not playerUpdates[k2] then
						playerUpdates[k2] = {usesIntellect = self:UsesIntellect(v2.class, v2.currentSpec)}
					else
						playerUpdates[k2].usesIntellect = self:UsesIntellect(v2.class, v2.currentSpec)
					end
				end
				-- If itemLevel, currentSpec, or weaponType has changed, update scale
				if v2.itemLevel and v2.currentSpec and v2.weaponType and ((not v1.itemLevel) or (not v1.currentSpec) or (not v1.weaponType) or (v1.itemLevel and v2.itemLevel ~= v1.itemLevel) or
					(v1.currentSpec and v2.currentSpec ~= v1.currentSpec) or (v1.weaponType and v2.weaponType ~= v1.weaponType)) then
					if not playerUpdates[k2] then
						playerUpdates[k2] = {scale = self:SelectBestScale(k2)}
					else
						playerUpdates[k2].scale = self:SelectBestScale(k2)
					end
				end
				-- If the scale has changed, update itemScores
				if v2.scale and ((not v1.scale) or v1.scale ~= v2.scale) then
					for slotName, playerInfo in pairs(self.db.char.slotPlayerItem) do
						for playerName, itemInfo in pairs(self.db.char.slotPlayerItem[slotName]) do
							if playerName == k2 then
								self:DBUpsert(self.db.char.slotPlayerItem[slotName][playerName], {itemScore = self:GetItemScore(itemInfo.link, playerName)})
							end
						end
					end
				end
			end
		end
	end
	self:DBDelete(self.db.char.playerIDs, idDelete)
	self:DBUpsert(self.db.char.playerIDs, idUpsert)
	self:DBUpdate(obj, playerUpdates)
	return deleted, inserted
end

function MasterLoot:TriggerSlotPlayerItemAfterInsert(obj, deleted, inserted)
	local playerUpdates = {}
	local itemUpdates = {}
	for k, v in pairs(inserted) do
		-- Update itemScore and player itemLevel
		if v.link and v.level then
			if self.db.char.player[k].usesIntellect ~= nil and self.db.char.player[k].scale then
				if not itemUpdates[k] then
					itemUpdates[k] = {itemScore = self:GetItemScore(v.link, k)}
				else
					itemUpdates.itemScore = self:GetItemScore(v.link, k)
				end
			end
			if not playerUpdates[k] then
				playerUpdates[k] = {itemLevel = self:GetItemLevel(k)}
			else
				playerUpdates[k].itemLevel = self:GetItemLevel(k)
			end
		end
	end
	self:DBUpsert(obj, itemUpdates)
	self:DBUpsert(self.db.char.player, playerUpdates)
	return deleted, inserted
end

function MasterLoot:TriggerSlotPlayerItemAfterUpdate(obj, deleted, inserted)
	local playerUpdates = {}
	local itemUpdates = {}
	for k1, v1 in pairs(deleted) do
		for k2, v2 in pairs(inserted) do
			if k1 == k2 then
				-- If link or level has changed, update itemScore and player itemLevel
				if v2.link and v2.level and self.db.char.player[k2].usesIntellect ~= nil and self.db.char.player[k2].scale and (((not v1.link) or (not v1.level)) or ((v1.link and v1.link ~= v2.link) or
					(v1.level and v1.level ~= v2.level))) then
					if not itemUpdates[k2] then
						itemUpdates[k2] = {itemScore = self:GetItemScore(v2.link, k2)}
					else
						itemUpdates[k2].itemScore = self:GetItemScore(v2.link, k2)
					end
					if not playerUpdates[k2] then
						playerUpdates[k2] = {itemLevel = self:GetItemLevel(k2)}
					else
						playerUpdates[k2].itemLevel = self:GetItemLevel(k2)
					end
				end
			end
		end
	end
	self:DBUpsert(obj, itemUpdates)
	self:DBUpsert(self.db.char.player, playerUpdates)
	return deleted, inserted
end

function MasterLoot:TriggerSlotPlayerItemAfterDelete(obj, deleted, inserted)
	local playerUpdates = {}
	for k, v in pairs(deleted) do
		-- Update player itemLevel
		if not playerUpdates[k] then
			playerUpdates[k] = {itemLevel = self:GetItemLevel(k)}
		else
			playerUpdates[k].itemLevel = self:GetItemLevel(k)
		end
	end
	self:DBUpdate(self.db.char.player, playerUpdates)
	return deleted, inserted
end

function MasterLoot:TriggerDropAfterInsert(obj, deleted, inserted)
	for key, value in pairs(inserted) do
		if value.view then
			lootButton = self:CreateLootButton(value)
			self.PlayerUpgradeFrame.children[1].children[1]:AddChild(lootButton)
			self.PlayerUpgradeFrame.children[1].children[1]:DoLayout()
			self.PlayerUpgradeFrame.frame:Show()
		end
	end
	return deleted, inserted
end

function MasterLoot:TriggerDropAfterUpdate(obj, deleted, inserted)
	for k1, v1 in pairs(deleted) do
		for k2, v2 in pairs(inserted) do
			-- If view is being added then create lootButton
			if (not v1.view) and v2.view then
				lootButton = self:CreateLootButton(v2)
				self.PlayerUpgradeFrame.children[1].children[1]:AddChild(lootButton)
				self.PlayerUpgradeFrame.children[1].children[1]:DoLayout()
				self.PlayerUpgradeFrame.frame:Show()
			end
			-- If the winner is being added, reset expTime, add winner to dropItemInfo, and disable lootButton
			if (not v1.winner) and v2.winner then
				local expTime = time() + 7200
				self:DBUpsert(obj, {[k2] = {expTime = expTime}})
				if self.PlayerUpgradeFrame then
					for key, lootButton in pairs(self.PlayerUpgradeFrame.children[1].children[1].children) do
						local dropInfo = lootButton:GetUserData("dropInfo")
						if dropInfo and dropInfo.lootSourceGuid == v2.lootSourceGuid and dropInfo.lootSlotIndex == v2.lootSlotIndex then
							lootButton:SetDisabled(true)
							local dropItemInfo = self.PlayerUpgradeFrame.children[2].children[1]
							if dropItemInfo and dropItemInfo:GetUserData("lootSourceGuid") == v2.lootSourceGuid and dropItemInfo:GetUserData("lootSlotIndex") == v2.lootSlotIndex then
								dropItemInfo:SetWinnerText(dropInfo.winner, dropInfo.class)
							end
						end
					end
				end
			end
		end
	end
end

-- -----------------------------------
-- Event Handlers
-- -----------------------------------

function MasterLoot:GROUP_ROSTER_UPDATE(event)
	self:ScheduleTimer("UpdateAllRoster", 1.5)
end

function MasterLoot:PARTY_LOOT_METHOD_CHANGED(event)
	self.lootMethod, self.masterLooterPartyID, self.masterLooterRaidID = GetLootMethod()
	self.lootThreshold = GetLootThreshold()
end

function MasterLoot:INSPECT_READY(event, guid)
	-- Look for a cancelled timer to determine if we are inspecting
	local name = self.db.char.playerGuids[guid]
	for key, value in pairs(self.timers.inspection) do
		if self:TimeLeft(value) == 0 then
			-- if we are inspecting, then requeue the player we were inspecting because we have interrupted the inspection
			self:NotifyInspect(select(6, GetPlayerInfoByGUID(key)))
			self.timers.inspection[key] = nil
		end
	end
	if self.timers.inspection[guid] then self:CancelTimer(self.timers.inspection[guid]) end
	if name and CanInspect(name) then		
		-- Get the player's current spec and add it to the table
		local specID = GetInspectSpecialization(name)
		local role = GetSpecializationRoleByID(specID)
		local currentSpec = select(2, GetSpecializationInfoByID(specID))
		if currentSpec then 
			self:DBUpsert(self.db.char.player, {[name] = {currentSpec = currentSpec, role = role}})
			self:UNIT_INVENTORY_CHANGED({[name] = 0})
		else
			self:NotifyInspect(name)
		end
	elseif name then
		self:NotifyInspect(name)
	end
	self.timers.inspection[guid] = nil
	self.inspection.status = "idle"
	self:NotifyInspect()
end

function MasterLoot:GET_ITEM_INFO_RECEIVED(event, itemID)
	-- Look through slotPlayerItem and find players who have the item equipped
	for slot, player in pairs(self.db.char.slotPlayerItem) do
		for playerName, value in pairs(self.db.char.slotPlayerItem[slot]) do
			if value.link and not value.itemScore then
				if idNum == self:GetItemId(value.link) then
					self:DBUpsert(self.db.char.slotPlayerItem[slot][playerName], {level = select(4, GetItemInfo(value.link))})
				end
			end
		end
	end
	-- Look through drop and if there is a match, get the item info
	local dropDeleted = {}
	local dropUpdated = {}
	for id, dropInfo in pairs(self.db.char.drop) do
		if idNum == self:GetItemId(dropInfo.link) then
			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = {GetItemInfo(dropInfo.link)}
			-- if the item is not weapon or armor, remove the drop
			if not(itemType == "Weapon" or itemType == "Armor") then
				for index, lootButton in pairs(self.PlayerUpgradeFrame.children[1].children[1].children) do
					if lootButton:GetUserData("dropInfo").lootSourceGuid == dropInfo.lootSourceGuid and lootButton:GetUserData("dropInfo").lootSlotIndex == dropInfo.lootSlotIndex then
						local dropItemInfo = self.PlayerUpgradeFrame.children[2]
						if dropItemInfo.children[1]:GetUserData("guid") == dropInfo.lootSourceGuid and dropItemInfo.children[1]:GetUserData("lootSlotIndex") == dropInfo.lootSlotIndex then
							if #dropItemInfo.children > 0 then dropItemInfo:ReleaseChildren() end
							if #self.PlayerUpgradeFrame.children[3].children[2].children[1].children > 0 then self.PlayerUpgradeFrame.children[3].children[2].children[1]:ReleaseChildren() end
						end
						for key, child in pairs(lootButton.parent.children) do
							if child == lootButton then
								tremove(lootButton.parent.children, key)
								lootButton.parent:DoLayout()
								lootButton:Release()
							end
						end
					end
				end
				dropDeleted[id] = dropInfo
			-- If we're keeping the drop, then create the view and add the name
			else
				dropUpdated[id] = {name = itemName, view = self:CreatePlayerUpgradeView(dropInfo.link)}
			end
		end
	end
	self:DBDelete(self.db.char.drop, dropDeleted)
	self:DBUpsert(self.db.char.drop, dropUpdated)
end

function MasterLoot:UNIT_INVENTORY_CHANGED(tUnit)
	for unit, numEvents in pairs(tUnit) do
		local name = GetUnitName(unit, true)
		local gearScore = 0 -- Sum of item levels of all equipped gear
		local numEquippedSlots = 0 -- Number of slots to use for avg equipped item level calculation
		local slot = {
			head = GetInventorySlotInfo("HeadSlot"),
			neck = GetInventorySlotInfo("NeckSlot"),
			shoulder = GetInventorySlotInfo("ShoulderSlot"),
			back = GetInventorySlotInfo("BackSlot"),
			chest = GetInventorySlotInfo("ChestSlot"),
			wrist = GetInventorySlotInfo("WristSlot"),
			hands = GetInventorySlotInfo("HandsSlot"),
			waist = GetInventorySlotInfo("WaistSlot"),
			legs = GetInventorySlotInfo("LegsSlot"),
			feet = GetInventorySlotInfo("FeetSlot"),
			finger0 = GetInventorySlotInfo("Finger0Slot"),
			finger1 = GetInventorySlotInfo("Finger1Slot"),
			trinket0 = GetInventorySlotInfo("Trinket0Slot"),
			trinket1 = GetInventorySlotInfo("Trinket1Slot"),
			mainHand = GetInventorySlotInfo("MainHandSlot"),
			secondaryHand = GetInventorySlotInfo("SecondaryHandSlot")
		}
		
		-- For each gear slot, get the item info and store in DB
		for slot, slotID in pairs(slot) do
			local inventoryItemLink = GetInventoryItemLink(name, slotID)
			if inventoryItemLink then
				local link = self:UnenchantItemLink(inventoryItemLink)
				local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link)
				if itemName then
					-- Add weapon type to player
					if slot == "mainHand" then
						if itemEquipLoc == "INVTYPE_WEAPON" or ((itemEquipLoc == "INVTYPE_RANGEDRIGHT" or itemEquipLoc == "INVTYPE_RANGED") and itemSubType == "Wands") or
							itemEquipLoc == "INVTYPE_WEAPONMAINHAND" then
							self:DBUpsert(self.db.char.player, {[name] = {weaponType = "INVTYPE_WEAPONMAINHAND"}})
						else
							self:DBUpsert(self.db.char.player, {[name] = {weaponType = "INVTYPE_2HWEAPON"}})
						end
					end
				end
				-- Add to slotPlayerItem
				self:DBUpsert(self.db.char.slotPlayerItem[slot], {[name] = {link = link, level = itemLevel}})
			elseif self.db.char.slotPlayerItem[slot] and self.db.char.slotPlayerItem[slot][name] and numEvents > 0 then
				self:DBDelete(self.db.char.slotPlayerItem[slot], {[name] = {}})
			end	
		end
		-- If items were missed, reinspect
		for slot, playerInfo in pairs(self.db.char.slotPlayerItem) do
			for playerName, itemInfo in pairs(self.db.char.slotPlayerItem[slot]) do
				if playerName == name then numEquippedSlots = numEquippedSlots + 1 end
			end
		end
		if (self.db.char.player[name].weaponType == "INVTYPE_2HWEAPON" and numEquippedSlots < 15) or (self.db.char.player[name].weaponType ~= "INVTYPE_2HWEAPON" and numEquippedSlots < 16) then
			self:ScheduleTimer("NotifyInspect", 10, name)
		end
	end
end

function MasterLoot:LOOT_OPENED(event, arg1)
	-- Insert a delay to allow loot info to populate
	self:ScheduleTimer("ScanLoot", 0.3, GetUnitName("target", false))
end

function MasterLoot:CHAT_MSG_LOOT(event, message)
	-- Parse message for potential winner and link
	local winner, link = strmatch(message, "(%a+%-?%a*)%sreceives?%sloot:%s(.*)|h|rx?%d*%.")
	if link then
		-- Get the proper name for the winner
		if winner == "You" then
			winner = self.playerName
		end
		-- If the link is in the drop table and it has no winner, then assign winner as the winner
		local found = false
		for id, dropInfo in pairs(self.db.char.drop) do
			if dropInfo.link == (link .. "|h|r") and dropInfo.winner == nil and (not found) then
				found = true
				self:DBUpsert(self.db.char.drop, {[id] = {winner = winner, class = self.db.char.player[winner].class}})
			end
		end
	end
end

-- -----------------------------------
-- Functions
-- -----------------------------------

function MasterLoot:CacheAllItems()
	for slot, player in pairs(self.db.char.slotPlayerItem) do
		for playerName, value in pairs(self.db.char.slotPlayerItem[slot]) do
			GetItemInfo(value.link)
		end
	end
	-- Look through drop and if there is a match, get the item info
	for id, value in pairs(self.db.char.drop) do
		GetItemInfo(value.link)
	end
end

function MasterLoot:round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function MasterLoot:NotifyInspect(unit, restart)
	local name
	if not self.inspection then self.inspection = {} end
	if not self.inspection.Q then self.inspection.Q = {} end
	if restart or (not self.inspection.status)  then self.inspection.status = "idle" end
	if unit then
		name = GetUnitName(unit, true)
		local found = false
		for key, value in pairs(self.inspection.Q) do
			if value == name then found = true end
		end
		if not found then tinsert(self.inspection.Q, name) end
	end
	if self.inspection.status ~= "stopped" then
		self.inspection.status = "running"
		-- If there are requests in queue, process the next request
		if #self.inspection.Q > 0 then
			if (not self.inspection.timer) or self:TimeLeft(self.inspection.timer) == 0 then
				self.inspection.numRequests = 0
				self.inspection.timer = self:ScheduleRepeatingTimer(function () self.inspection.numRequests = 0 end, 2)
			end
			-- If we have already reached our limit for this interval, then wait for bandwidth to send one more and try again
			if self.inspection.numRequests == 1 then
				self:ScheduleTimer("NotifyInspect", 2)
			else
			-- We're not at the limit so go ahead
				self.inspection.numRequests = self.inspection.numRequests + 1
				if CanInspect(self.inspection.Q[1]) and ((not self.timers.inspection[guid]) or self:TimeLeft(self.timers.inspection[guid] == 0)) then
					local guid = UnitGUID(self.inspection.Q[1])
					self.inspection.status = "stopped"
					self.hooks.NotifyInspect(self.inspection.Q[1])
					-- schedule a retry timer in case of timeout
					self.timers.inspection[guid] = self:ScheduleTimer("NotifyInspect", self:round(2 + ((select(4, GetNetStats()) * 2) / 1000)), name, true)
				elseif self.db.char.player[name] then
					-- move to the back of the queue
					tinsert(self.inspection.Q, name)
				end
				tremove(self.inspection.Q, 1)
				-- Continue the process until no requests are left
				self:NotifyInspect()
			end
		else
			self.inspection.status = "idle"
		end
	end
end

function MasterLoot:UsesIntellect(class, currentSpec)
	assert(class, "Error in UsesIntellect: class not specified.")
	assert(currentSpec, "Error in UsesIntellect: currentSpec not specified.")
	assert(type(class) == "string", "Error in UsesIntellect: class not of type string.")
	assert(type(currentSpec) == "string", "Error in UsesIntellect: currentSpec not of type string.")
	local specs = {
		["DRUID"] = {"Balance", "Restoration"},
		["MAGE"] = {"Arcane", "Fire", "Frost"},
		["MONK"] = {"Mistweaver"},
		["PALADIN"] = {"Holy"},
		["PRIEST"] = {"Discipline", "Holy", "Shadow"},
		["SHAMAN"] = {"Elemental", "Restoration"},
		["WARLOCK"] = {"Affliction", "Demonology", "Destruction"}
	}
	
	if specs[class] then
		for key, spec in pairs(specs[class]) do
			if spec == currentSpec then
				return true
			end
		end
	end
	
	return false
end

function MasterLoot:translateStats(itemLink, playerName)
	-- Translates an item stat table into stats active for the specified target player
	local translated = {}
	local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(itemLink)
	local stats = GetItemStats(itemLink)
	-- Translate Resistance0 stats to Armor or BonusArmor
	for stat, value in pairs(stats) do
		if stat == "RESISTANCE0_NAME" then
			if itemEquipLoc == "INVTYPE_NECK" or itemEquipLoc == "INVTYPE_FINGER" or itemEquipLoc == "INVTYPE_TRINKET" or itemEquipLoc == "INVTYPE_CLOAK" then
				translated["BonusArmor"] = value
			else
				translated["Armor"] = value
			end
		else
			translated[stat] = value
		end
	end
	-- Look for intellect, strength, or agility in the stat table
	if itemType == "Armor" and (itemEquipLoc == "INVTYPE_HEAD" or itemEquipLoc == "INVTYPE_SHOULDER" or itemEquipLoc == "INVTYPE_CHEST" or itemEquipLoc == "INVTYPE_ROBE" or
		itemEquipLoc == "INVTYPE_WRIST" or itemEquipLoc == "INVTYPE_HAND" or itemEquipLoc == "INVTYPE_WAIST" or itemEquipLoc == "INVTYPE_LEGS" or itemEquipLoc == "INVTYPE_FEET") then
		for stat, value in pairs(translated) do
			-- If strength or agility is found on plate, mail, or leather and the target player uses intellect, then translate the stat to intellect
			if (stat == "ITEM_MOD_AGILITY_SHORT" or stat == "ITEM_MOD_STRENGTH_SHORT") and (itemSubType == "Plate" or itemSubType == "Mail" or itemSubType == "Leather") and
				self.db.char.player[playerName].usesIntellect then
				translated["ITEM_MOD_INTELLECT_SHORT"] = value
				translated[stat] = nil
			-- If intellect is found and the target player does not use intellect, then translate the stat
			elseif stat == "ITEM_MOD_INTELLECT_SHORT" and not self.db.char.player[playerName].usesIntellect then
				-- If the armor subType is plate, then translate the stat into strength
				if itemSubType == "Plate" then
					translated["ITEM_MOD_STRENGTH_SHORT"] = value
					translated[stat] = nil
				-- If the armor subType is mail or leather, then translate the stat into agility
				elseif itemSubType == "Mail" or itemSubType == "Leather" then
					translated["ITEM_MOD_AGILITY_SHORT"] = value
					translated[stat] = nil
				end
			end
		end
	end
	
	-- Other item types/subtypes do not translate
	return translated
end

function MasterLoot:UnenchantItemLink(ItemLink)
	assert(ItemLink, "Error in UnenchantItemLink: item link not specified.")
	assert(type(ItemLink) == "string", "Error in UnenchantItemLink: ItemLink not of type string.")
	local ItemID, EnchantID, GemID1, GemID2, GemID3, GemID4, SuffixID, MoreInfo, ViewAtLevel, UpgradeLevel, Difficulty, NumBonusIDs, BonusIDs = strmatch(ItemLink, "item:(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):?([%d:]*)")
	
	if Difficulty then
		-- Return a new link.
		local newItemID = "item:" .. ItemID .. ":0:0:0:0:0:" .. SuffixID .. ":" .. MoreInfo .. ":" .. ViewAtLevel .. ":" .. UpgradeLevel .. ":" .. Difficulty .. ":" .. NumBonusIDs
		if BonusIDs then
			for i = 1, tonumber(NumBonusIDs), 1 do
				local id, last = strsplit(":", BonusIDs, 2)
				BonusIDs = last
				newItemID = newItemID .. ":" .. id
			end
		end
		return gsub(ItemLink, "item[%d:]+", newItemID)
	end
end

function MasterLoot:GetItemLevel(playerName)
	local gearScore = 0
	local numEquippedSlots = 0
	for slot, playerInfo in pairs(self.db.char.slotPlayerItem) do
		for name, itemInfo in pairs(self.db.char.slotPlayerItem[slot]) do
			if name == playerName then
				numEquippedSlots = numEquippedSlots + 1
				if itemInfo.level then gearScore = gearScore + itemInfo.level end
			end
		end
	end
	if not (numEquippedSlots == 0) then return gearScore / numEquippedSlots end
end

function MasterLoot:SelectBestScale(playerName)
	local candidateScales = {}
	local useItemLevel = false -- If at least one of the candidate scales has an item level, then this will be set to true because we prefer to use scales customized to item level
	local useWeaponType = false -- If at least one of the candidate scales has a weapon type, then this will be set to true because we prefer to use scales customized to weapon type
	local bestItemLevel -- Stores the item level that is closest to the player's item level when using item level scales
	-- Look for scales for the player's class and spec
	assert(self.db.char.player[playerName].class, "Cannot determine scale for " .. playerName .. " because player class is missing.")
	assert(self.db.char.player[playerName].currentSpec, "Cannot determine scale for " .. playerName .. " because player's current spec is missing.")
	for name, value in pairs(self.scale) do
		if self.db.char.player[playerName].class == value.class and self.db.char.player[playerName].currentSpec == value.SpecName then
			candidateScales[name] = value
			-- If the scale is customized for item level, set the appropriate flag
			if value.itemLevel then useItemLevel = true end
		end
	end
	-- If we are using scales customized for item level, find the best match(es)
	if useItemLevel then
		for scaleName, value in pairs(candidateScales) do
			-- Filter out scales with no item level
			if not value.itemLevel then
				candidateScales[scaleName] = nil
			-- Calculate the best item level (the one closest to the player's item level and below the player's item level if such a scale exists)
			-- If there is no best item level established, then update the best item level
			-- OR if the best item level > the player's item level and the scale's item level <= the best item level, update the best item level
			-- OR if the best item level <= the player's item level and the scale's item level <= the player's item level and closer to the player's item level, update the best item level
			--O R if the player's item level is nil and the scale's item level < best item level
			elseif (not bestItemLevel) or (bestItemLevel > self.db.char.player[playerName].itemLevel and value.itemLevel < bestItemLevel) or
				(bestItemLevel <= self.db.char.player[playerName].itemLevel and value.itemLevel <= self.db.char.player[playerName].itemLevel and value.itemLevel > bestItemLevel) or
				((not self.db.char.player[playerName].itemLevel) and value.itemLevel < bestItemLevel) then 
				bestItemLevel = value.itemLevel
			end
		end
		-- Filter out scales that do not match the best item level
		for scaleName, value in pairs(candidateScales) do
			if value.itemLevel ~= bestItemLevel then
				candidateScales[scaleName] = nil
			end
		end
	end
	-- If any of the remaining scales specify a weapon type matching the player's weapon type, then set the weapontype flag
	for scaleName, value in pairs(candidateScales) do
		if value.weaponType and self.db.char.player[playerName].weaponType == value.weaponType then useWeaponType = true end
	end
	-- If we are using scales customized for weapon type, filter out ones that do not specify a weapon type
	if useWeaponType then
		for scaleName, value in pairs(candidateScales) do
			if not value.weaponType then candidateScales[scaleName] = nil end
		end
	end
	-- There should be only one remaining scale so return it;
	local numScales = 0
	for scaleName, value in pairs(candidateScales) do
		numScales = numScales + 1
	end
	if (not numScales == 1) then
		error("Unable to determine best scale for " .. playerName .. " because " .. numScales .. " candidate scales exist.")
	end
	for scaleName, value in pairs(candidateScales) do return scaleName end
end

function MasterLoot:GetItemScore(itemLink, playerName)
	--Get the translated item stats table for player
	local tStats = self:translateStats(itemLink, playerName)
	--Iterate throught the stats and add the weighted stat score to the total item score
	local itemScore = 0
	for statName, statValue in pairs (tStats) do
		if self.db.char.player[playerName].scale and self.scale[self.db.char.player[playerName].scale][statName] then
			itemScore = itemScore + (statValue * self.scale[self.db.char.player[playerName].scale][statName])
		end
	end
	return itemScore
end

function MasterLoot:UpdateAllRoster()
	-- update players.
	local deleted = {}
	local inserted = {}
	local retry = false
	if IsInRaid() then
		for i = 1, GetNumGroupMembers() do
			local name, rank, subgroup, level, class, className = GetRaidRosterInfo(i)
			local id = "raid" .. i
			local guid = UnitGUID(id)
			if name then
				inserted[name] = {class = className, id = id, guid = guid}
			else
				retry = true
			end
		end
	else
		for i = 0, GetNumSubgroupMembers() do
			local id
			if (i == 0) then
				id = "player"
			else
				id = "party"..i
			end
			local name = GetUnitName(id, true)
			local className = select(2, UnitClass(id))
			local guid = UnitGUID(id)
			if name then
				inserted[name] = {guid = guid, class = select(2, UnitClass(id)), id = id}
			else
				retry = true
			end
		end
	end
	for playerName, playerInfo in pairs(self.db.char.player) do
		local found = false
		for key, value in pairs(inserted) do
			if playerName == key then found = true end
		end
		if not found then deleted[playerName] = playerInfo end
	end
	self:DBDelete(self.db.char.player, deleted)
	self:DBUpsert(self.db.char.player, inserted)
	if retry then self:GROUP_ROSTER_UPDATE() end
end

function MasterLoot:GetItemId(item)
	if type(item) == "number" then
		return item
	else
		return tonumber(string.match(item, "item:(%d+):"))
	end
end

function MasterLoot:IsArmorUsable(itemSubtype, playerClass)
	-- if the item is cloth or miscellaneous, return true
	if itemSubtype == "Cloak" or itemSubtype == "Miscellaneous" then
		return true
	else
		local prof = {
			Cloth = {"MAGE", "PRIEST", "WARLOCK"},
			Leather = {"DRUID", "MONK", "ROGUE"},
			Mail = {"HUNTER", "SHAMAN"},
			Plate = {"DEATHKNIGHT", "PALADIN", "WARRIOR"},
			Shields = {"PALADIN", "SHAMAN", "WARRIOR"}
		}
		for key, class in pairs(prof[itemSubtype]) do
			if class == playerClass then
				return true
			else
				return false
			end
		end
	end
end

function MasterLoot:IsWeaponUsable(itemSubtype, playerClass)
	local prof = {
		Bows = {"HUNTER", "WARRIOR"},
		Crossbows = {"HUNTER", "WARRIOR"},
		Daggers = {"DRUID", "HUNTER", "MAGE", "PRIEST", "ROGUE", "SHAMAN", "WARLOCK", "WARRIOR"},
		Guns = {"HUNTER", "WARRIOR"},
		["Fist Weapons"] = {"DRUID", "HUNTER", "MONK", "ROGUE", "SHAMAN", "WARRIOR"},
		["One-Handed Axes"] = {"MONK", "ROGUE"},
		["One-Handed Maces"] = {"MONK", "PRIEST", "ROGUE"},
		["One-Handed Swords"] = {"MONK", "ROGUE"},
		Polearms = {"DEATHKNIGHT", "DRUID", "HUNTER", "MONK", "PALADIN", "WARRIOR"},
		Staves = {"DRUID", "HUNTER", "MAGE", "MONK", "PRIEST", "SHAMAN", "WARLOCK", "WARRIOR"},
		Thrown = {"ROGUE", "WARRIOR"},
		["Two-Handed Axes"] = {"DEATHKNIGHT", "HUNTER", "PALADIN", "SHAMAN", "WARRIOR"},
		["Two-Handed Maces"] = {"DEATHKNIGHT", "DRUID", "PALADIN", "SHAMAN", "WARRIOR"},
		["Two-Handed Swords"] = {"DEATHKNIGHT", "HUNTER", "MAGE", "PALADIN", "WARLOCK", "WARRIOR"},
		Wands = {"MAGE", "PRIEST", "WARLOCK"}
	}
	for key, class in pairs(prof[itemSubtype]) do
		if class == playerClass then
			return true
		else
			return false
		end
	end
end

function MasterLoot:IsItemUsable(itemLink, playerName)
	local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(itemLink)
	local stats = self:translateStats(itemLink, playerName)
	-- If the player currently has the item equipped, then return false
	for slot, playerInfo in pairs(self.db.char.slotPlayerItem) do
		for name, itemInfo in pairs(self.db.char.slotPlayerItem[slot]) do
			if itemLink == itemInfo.link and Name == playerName then
				return false
			end
		end
	end
	-- If bonus armor or spirit, only useful for tanks or healers
	for stat, value in pairs(stats) do
		if stat == "BonusArmor" then
			if self.db.char.player[playerName].role == "TANK" then
				return true
			else
				return false
			end
		elseif stat == "ITEM_MOD_SPIRIT_SHORT" then
			if self.db.char.player[playerName].role == "HEALER" then
				return true
			else
				return false
			end
		-- If intellect, only useful for intellect users
		elseif stat == "ITEM_MOD_INTELLECT_SHORT" and not self.db.char.player[playerName].usesIntellect then
			return false
		-- If strength or agility, only useful for non-intellect users
		elseif (stat == "ITEM_MOD_STRENGTH_SHORT" or stat == "ITEM_MOD_AGILITY_SHORT") and self.db.char.player[playerName].usesIntellect then
			return false
		end
	end
	-- if the item is armor, call IsArmorUsable
	if itemType == "Armor" then
		return self:IsArmorUsable(itemSubType, self.db.char.player[playerName].class)
	-- if the item is a weapon, call IsWeaponUsable
	elseif itemType == "Weapon" then
		return self:IsWeaponUsable(itemSubType, self.db.char.player[playerName].class)
	else
		return false
	end
end

function MasterLoot:CreatePlayerUpgradeView(itemLink)
	local itemEquipLoc = select(9, GetItemInfo(itemLink))
	if itemEquipLoc then
		local equipLocMap = {}
		equipLocMap["INVTYPE_HEAD"] =  {"head"}
		equipLocMap["INVTYPE_NECK"] =  {"neck"}
		equipLocMap["INVTYPE_SHOULDER"] =  {"shoulder"}
		equipLocMap["INVTYPE_CHEST"] =  {"chest"}
		equipLocMap["INVTYPE_ROBE"] =  {"chest"}
		equipLocMap["INVTYPE_WAIST"] =  {"waist"}
		equipLocMap["INVTYPE_LEGS"] =  {"legs"}
		equipLocMap["INVTYPE_FEET"] =  {"feet"}
		equipLocMap["INVTYPE_WRIST"] =  {"wrist"}
		equipLocMap["INVTYPE_HAND"] =  {"hands"}
		equipLocMap["INVTYPE_FINGER"] =  {"finger0", "finger1"}
		equipLocMap["INVTYPE_TRINKET"] =  {"trinket0", "trinket1"}
		equipLocMap["INVTYPE_CLOAK"] =  {"back"}
		equipLocMap["INVTYPE_WEAPON"] =  {"mainHand", "secondaryHand"}
		equipLocMap["INVTYPE_SHIELD"] =  {"secondaryHand"}
		equipLocMap["INVTYPE_2HWEAPON"] =  {"mainHand"}
		equipLocMap["INVTYPE_WEAPONMAINHAND"] =  {"mainHand"}
		equipLocMap["INVTYPE_WEAPONOFFHAND"] =  {"secondaryHand"}
		equipLocMap["INVTYPE_HOLDABLE"] =  {"secondaryHand"}
		equipLocMap["INVTYPE_RANGED"] =  {"mainHand"}
		equipLocMap["INVTYPE_RANGEDRIGHT"] = {"mainHand"}
		-- TODO: For weapons, get both hands
		local equipSlot = equipLocMap[itemEquipLoc]
		local view = {}
		
		-- Go through the entire roster
		for name, info in pairs(self.db.char.player) do
			-- If the dropped item is usable for the player, try to find the player in the appropriate slot records in slotPlayerItem
			if self:IsItemUsable(itemLink, name) then
				for slot, playerInfo in pairs(self.db.char.slotPlayerItem) do
					for key, slotName in pairs(equipSlot) do
						if slot == slotName then
							local found = false
							-- iterate through self.db.char.slotPlayerItem.[equipSlot] and add players/items for which the drop is usable
							for playerName, itemInfo in pairs (self.db.char.slotPlayerItem[slot]) do
								if name == playerName then
									view[slot .. "|" .. playerName] = {slot = slot, player = playerName, class = info.class, equippedItemLink = itemInfo.link, equippedItemScore = itemInfo.itemScore,
										visible = true}
									view[slot .. "|" .. playerName].droppedItemScore = self:GetItemScore(itemLink, playerName)
									if itemInfo.itemScore and itemInfo.itemScore ~= 0 then
										view[slot .. "|" .. playerName].upgradePct = ((view[slot .. "|" .. playerName].droppedItemScore - itemInfo.itemScore) / itemInfo.itemScore) * 100
									elseif not itemInfo.itemScore then
										view[slot .. "|" .. playerName].itemScoreMissing = true
									end
									found = true
								end
							end
							-- Also add records for players that do not have equipped item info for the given slot
							if not found then
								view[slot .. "|" .. name] = {slot = slot, player = name, class = info.class, equippedItemSlot = slot, visible = true, itemLinkMissing = true}
							end
						end
					end
				end
			end
		end
		-- Go back through the view and perform some cleanup activities
		for key, value in pairs(view) do
			-- If the player has a 2H weapon currently equipped, hide the secondary hand
			if self.db.char.player[value.player].weaponType == "INVTYPE_2HWEAPON" and value.equippedItemSlot == "secondaryHand" then
				view[key].visible = false
			end
			-- If the player has a main hand weapon equipped, then both hands and their weighted stat scores are displayed
			-- For main hand weapon or off hand/1H drops
			if value.equippedItemSlot == "mainHand" then
				-- If the player has a 2H weapon equipped, the 2H weapon and weighted stat score  is displayed with a warning
				if self.db.char.player[value.player].weaponType == "INVTYPE_2HWEAPON" then
					weaponTypeWarning = true
				-- If the player has a main hand weapon equipped, search for an offhand weapon and display the lower value of the two if one exists
				else
					for key2, value2 in pairs (view) do
						if value2.equippedItemSlot == "secondaryHand" and value2.equippedItemLink and (select(9, GetItemInfo(value2.equippedItemLink)) == "INVTYPE_WEAPONOFFHAND" or
							select(9, GetItemInfo(value2.equippedItemLink)) == "INVTYPE_WEAPON") and value.equippedItemScore < value2.equippedItemScore then
							view[key2].visible = false
						elseif value2.equippedItemSlot == "secondaryHand" and value2.equippedItemLink and (select(9, GetItemInfo(value2.equippedItemLink)) == "INVTYPE_WEAPONOFFHAND" or
							select(9, GetItemInfo(value2.equippedItemLink)) == "INVTYPE_WEAPON") then
							view[key].visible = false
						--  If an offhand weapon does not exist, hide the offhand slot
						elseif value2.equippedItemSlot == "secondaryHand" then
							view[key2].visible = false
						end
					end
				end
			elseif itemEquipLoc == "INVTYPE_FINGER" and value.equippedItemSlot == "finger0" then
			 for key2, value2 in pairs(view) do
				if value.equippedItemSlot == "finger1" and value.equippedItemScore < value2.equippedItemScore then
					view[key2].visible = false
				elseif value.equippedItemSlot == "finger1" then
					view[key].visible = false
				end
			 end
			elseif itemEquipLoc == "INVTYPE_TRINKET" and value.equippedItemSlot == "trinket0" then
				if value.equippedItemSlot == "trinket1" and value.equippedItemScore < value2.equippedItemScore then
					view[key2].visible = false
				elseif value.equippedItemSlot == "trinket1" then
					view[key].visible = false
				end
			end
		end
		return view
	end
end

function MasterLoot:ScanLoot(sourceName)
	-- Create a loot drop event for armor and weapons at or above the loot threshold
	-- Event id = sourceguid, loot slot
	local drops = {}
	local looted = {}
	for key, dropInfo in pairs(self.db.char.drop) do
		if looted[dropInfo.lootSourceGuid] then
			looted[dropInfo.lootSourceGuid] = looted[dropInfo.lootSourceGuid] + 1
		else
			looted[dropInfo.lootSourceGuid] = 1
		end
	end
	for lootSlotIndex = 1, GetNumLootItems(), 1 do
		local guid, quant = GetLootSourceInfo(lootSlotIndex)
		local itemLink = GetLootSlotLink(lootSlotIndex)
		local lootIcon, lootName, lootQuantity, lootQuality, locked, isQuestItem, questID, isActive = GetLootSlotInfo(lootSlotIndex)
		if (not looted[guid]) and lootQuality >= self.lootThreshold then
			local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(itemLink)
			if itemType == "Weapon" or itemType == "Armor" then
				drops[guid .. "|" .. lootSlotIndex] = {link = itemLink, icon = lootIcon, name = itemName, rarity = lootQuality, lootSourceGuid = guid, lootSlotIndex = lootSlotIndex,
					sourceName = sourceName, expTime = time() + 7200, view = self:CreatePlayerUpgradeView(itemLink)}
			elseif not itemType then
				drops[guid .. "|" .. lootSlotIndex] = {link = itemLink, icon = lootIcon, rarity = lootQuality, lootSourceGuid = guid, lootSlotIndex = lootSlotIndex,
					sourceName = sourceName, expTime = time() + 7200}
			end
		end
	end
	self:DBUpsert(self.db.char.drop, drops)
end

function MasterLoot:GetReforgingText(link)
	assert(link, "Error in GetReforgingText: item link not specified.")
	assert(type(link) == "string", "Error in GetReforgingText: link not of type string.")
	local NumBonusIDs, BonusIDs = strmatch(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:(%d+):?([%d:]*)")
	if BonusIDs then
		local Heroic = false
		local Mythic = false
		local Warforged = false
		for i = 1, tonumber(NumBonusIDs), 1 do
			local id, last = strsplit(":", BonusIDs, 2)
			BonusIDs = last
			if tonumber(id) == 566 then
				Heroic =  true
			elseif tonumber(id) == 567 then
				Mythic = true
			elseif tonumber(id) == 560 or tonumber(id) == 561 or tonumber(id) == 562 then
				Warforged = true
			end
		end
		local text
		if Heroic then
			text = "Heroic"
		elseif Mythic then
			text = "Mythic"
		end
		if Warforged and not (Heroic or Mythic) then
			text = "Warforged"
		elseif Warforged then
			text = text .. " Warforged"
		end
		if text then return text end
	end
end

function MasterLoot:CreatePlayerUpgradeFrame()
	local frame = AceGUI:Create("Frame")
	frame.frame:SetMinResize(700, 200)
	frame:SetWidth(775)
	frame:SetTitle("Player Upgrade Comparison")
	frame:SetStatusText("Master Loot BETA")
	frame:SetCallback("OnClose", function(widget) widget.frame:Hide() end)
	frame:SetLayout("PlayerUpgradeLayout")

	local dropContainer = AceGUI:Create("SimpleInlineGroup")
	dropContainer:SetLayout("Fill")
	frame:AddChild(dropContainer)

	local dropContainerScrollFrame = AceGUI:Create("ScrollFrame")
	dropContainerScrollFrame:SetLayout("TightList")
	dropContainer:AddChild(dropContainerScrollFrame)
	
	local dropDetails = AceGUI:Create("SimpleGroup")
	dropDetails:SetLayout("Fill")
	frame:AddChild(dropDetails)
	
	local playerUpgradeViewGroup = AceGUI:Create("SimpleGroup")
	playerUpgradeViewGroup:SetLayout("PlayerUpgradeViewGroupLayout")
	frame:AddChild(playerUpgradeViewGroup)
	
	local PlayerUpgradeViewHeader = AceGUI:Create("PlayerUpgradeViewHeader")
	playerUpgradeViewGroup:AddChild(PlayerUpgradeViewHeader)	

	local playerUpgradeViewLinesGroup = AceGUI:Create("SimpleInlineGroup")
	playerUpgradeViewLinesGroup:SetLayout("Fill")
	playerUpgradeViewGroup:AddChild(playerUpgradeViewLinesGroup)

	local playerUpgradeViewScrollFrame = AceGUI:Create("ScrollFrame")
	playerUpgradeViewScrollFrame:SetLayout("TightList")
	playerUpgradeViewLinesGroup:AddChild(playerUpgradeViewScrollFrame)

	local first = true
	for id, dropInfo in pairs(self.db.char.drop) do
		if dropInfo.view then
			lootButton = self:CreateLootButton(dropInfo)
			dropContainerScrollFrame:AddChild(lootButton)
			if first then
				dropDetails:AddChild(self:CreateDropItemInfo(dropInfo))
				for key, value in pairs(self:CreatePlayerUpgradeViewLines(dropInfo.view)) do
					playerUpgradeViewScrollFrame:AddChild(value)
				end
				first = false
			end
		end
	end
	
	frame:DoLayout()
	dropDetails:DoLayout()
	playerUpgradeViewGroup:DoLayout()
	playerUpgradeViewLinesGroup:DoLayout()
	
	return frame
end

function MasterLoot:FormatTime(seconds)
	assert(seconds, "Error in FormatTime: seconds not specified.")
	assert(type(seconds) == "number", "Error in FormatTime: seconds is not of type number.")
	assert(seconds >= 0, "Error in FormatTime: seconds must be >= 0.")
	
	local h = math.floor(math.floor(seconds) / 3600)
	local m = math.floor((math.floor(seconds) % 3600) / 60)
	local hString = tostring(h)
	local mString
	
	if h > 0 and m < 10 then
		mString = "0" .. tostring(m)
	else
		mString = tostring(m)
	end

	if h > 0 then
		return hString .. ":" .. mString
	elseif m > 1 then
		return mString .. " minutes"
	else
		return "< 1 minute"
	end
end

function MasterLoot:UpdateLootButton(lootButton)
	local dropInfo = lootButton:GetUserData("dropInfo")
	-- If it's past the expiration time, delete the drop, cancel the timer, and release widgets
	if time() >= dropInfo.expTime then
		if lootButton:GetUserDataTable().timer then self:CancelTimer(lootButton:GetUserData("timer")) end
		local dropItemInfo = self.PlayerUpgradeFrame.children[2]
		if dropItemInfo.children[1] and dropItemInfo.children[1]:GetUserData("guid") == dropInfo.lootSourceGuid and dropItemInfo.children[1]:GetUserData("lootSlotIndex") == dropInfo.lootSlotIndex then
			 if #dropItemInfo.children > 0 then dropItemInfo:ReleaseChildren() end
			if #self.PlayerUpgradeFrame.children[3].children[2].children[1].children > 0 then self.PlayerUpgradeFrame.children[3].children[2].children[1]:ReleaseChildren() end
		end
		for key, child in pairs(lootButton.parent.children) do
			if child == lootButton then
				tremove(lootButton.parent.children, key)
				lootButton.parent:DoLayout()
				lootButton:Release()
			end
		end
		self:DBDelete(self.db.char.drop, {[dropInfo.lootSourceGuid .. "|" .. dropInfo.lootSlotIndex] = dropInfo})
	else
		lootButton:SetTimeText(self:FormatTime(dropInfo.expTime - time()) .. " remaining")
	end
end

function MasterLoot:CreateLootButton(dropInfo)
	local lootButton = AceGUI:Create("LootButton")
	lootButton:SetUserData("dropInfo", dropInfo)
	lootButton:SetImage(dropInfo.icon)
	lootButton:SetNameText(dropInfo.name)
	lootButton:SetColor(GetItemQualityColor(dropInfo.rarity))
	if dropInfo.sourceName then
		lootButton:SetSourceText(dropInfo.sourceName)
	else
		lootButton:SetSourceText("Unknown")
	end
	self:ScheduleTimer(function(lootButton)
		self:UpdateLootButton(lootButton)
		if
			(dropInfo.expTime - time()) > 0 then
			lootButton:SetUserData("timer", self:ScheduleRepeatingTimer("UpdateLootButton", 60, lootButton))
		else
			self:UpdateLootButton(lootButton)
		end
	end, (dropInfo.expTime - time()) % 60, lootButton)
	lootButton:SetTimeText(self:FormatTime(dropInfo.expTime - time()) .. " remaining")
	if dropInfo.winner then
		lootButton:SetDisabled(true)
	end
	
	lootButton:SetCallback("OnClick", function(lootButton)
		local dropDetails = lootButton.parent.parent.parent.children[2]
		local playerUpgradeViewScrollFrame = lootButton.parent.parent.parent.children[3].children[2].children[1]
		if #dropDetails.children > 0 then dropDetails:ReleaseChildren() end
		if #playerUpgradeViewScrollFrame.children > 0 then playerUpgradeViewScrollFrame:ReleaseChildren() end
		dropDetails:AddChild(self:CreateDropItemInfo(lootButton:GetUserData("dropInfo")))
		for key,value in pairs(self:CreatePlayerUpgradeViewLines(lootButton:GetUserData("dropInfo").view)) do
			playerUpgradeViewScrollFrame:AddChild(value)
		end
		dropDetails:DoLayout()
	end)
	return lootButton
end

function MasterLoot:CreateDropItemInfo(dropInfo)
	local dropItemInfo = AceGUI:Create("ItemInfo")
	dropItemInfo:SetUserData("guid", dropInfo.lootSourceGuid)
	dropItemInfo:SetUserData("lootSlotIndex", dropInfo.lootSlotIndex)
	dropItemInfo:SetImage(dropInfo.icon)
	dropItemInfo:SetNameText(dropInfo.name)
	dropItemInfo:SetColor(GetItemQualityColor(dropInfo.rarity))
	dropItemInfo:SetReforgingText(self:GetReforgingText(dropInfo.link))
	dropItemInfo:SetEquipLocText(_G[select(9, GetItemInfo(dropInfo.link))])
	if dropInfo.winner then
		dropItemInfo:SetWinnerText(dropInfo.winner, dropInfo.class)
	else
		dropItemInfo:SetWinnerText()
	end
	return dropItemInfo
end

function MasterLoot:CreatePlayerUpgradeViewLines(view)
	local lines = {}

	for key, viewLine in pairs(view) do
		if viewLine.visible then
			local line = AceGUI:Create("PlayerUpgradeViewLine")
			line:SetPlayerText(strmatch(viewLine.player, "([^%-]+)%-?.*"), viewLine.class)
			if not viewLine.itemLinkMissing then
				local itemName, _, itemRarity = GetItemInfo(viewLine.equippedItemLink)
				line:SetNameText(itemName, itemRarity)
				if not viewLine.itemScoreMissing then
					line:SetEquippedItemScoreText(viewLine.equippedItemScore)
					if viewLine.upgradePct then
						line:SetPctUpgradeText(tostring(self:round(viewLine.upgradePct, 2)))
					else
						line:SetPctUpgradeText()
					end
				else
					line:SetPctUpgradeText()
					line:SetEquippedItemScoreText("")
				end
			else
				line:SetNameText("Not Found", 0)
				line:SetPctUpgradeText()
				line:SetEquippedItemScoreText("")
			end

			line:SetDroppedItemScoreText(viewLine.droppedItemScore)

			tinsert(lines, line)
		end
	end
	return lines
end